"""
rtl_433 Kismet data source

(c) 2018 Mike Kershaw / Dragorn
Licensed under GPL2 or above

Supports both local usb rtlsdr devices via the rtl_433 binary

Sources are generated as rtl433-XYZ when multiple rtl radios are detected.

Accepts standard options:
    channel=freqMHz   (in mhz)
    channel=freqKHz   (in khz)
    channel=freq      (in raw hz to rtl_433)

    channels="a,b,c"  Pass hopping list to rtl433_bin
"""
from __future__ import print_function

import asyncio
import argparse
import ctypes
from datetime import datetime
import json
import os
import subprocess
import sys
import threading
import traceback
import time
import uuid

from . import kismetexternal

class KismetRtl433(object):
    def __init__(self):
        self.opts = {}

        self.opts['rtlbin'] = 'rtl_433'
        self.opts['channel'] = "433.920MHz"
        self.opts['gain'] = None
        self.opts['device'] = None
        self.opts['uuid'] = None
        self.opts['ppm'] = None
        self.opts['debug'] = None

        # The subprocess
        self.rtl_proc = None

        # The task so we can kill it during reconfigure
        self.rtl_task = None

        # Are we killing rtl because we're reconfiguring?
        self.rtl_reconfigure = False

        # We're usually not remote
        self.proberet = None

        # Do we have librtl?
        self.have_librtl = False

        self.driverid = "rtl433"
        # Use ctypes to load librtlsdr and probe for supported USB devices
        try:
            self.rtllib = ctypes.CDLL("librtlsdr.so.0")

            self.rtl_get_device_count = self.rtllib.rtlsdr_get_device_count

            self.rtl_get_device_name = self.rtllib.rtlsdr_get_device_name
            self.rtl_get_device_name.argtypes = [ctypes.c_int]
            self.rtl_get_device_name.restype = ctypes.c_char_p

            self.rtl_get_usb_strings = self.rtllib.rtlsdr_get_device_usb_strings
            self.rtl_get_usb_strings.argtypes = [ctypes.c_int, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p]

            self.rtl_get_index_by_serial = self.rtllib.rtlsdr_get_index_by_serial
            self.rtl_get_index_by_serial.argtypes = [ctypes.c_char_p]
            self.rtl_get_index_by_serial.restype = ctypes.c_int

            self.have_librtl = True
        except OSError:
            self.have_librtl = False

        parser = argparse.ArgumentParser(description='RTL433 to Kismet bridge - Creates a rtl433 data source on a Kismet server and passes JSON-based records from the rtl_433 binary',
                epilog='Requires the rtl_433 tool (install your distributions package or compile from https://github.com/merbanan/rtl_433)')

        parser = kismetexternal.ExternalInterface.common_getopt(parser)
        
        self.config = parser.parse_args()

        if not self.config.connect == None and self.config.source == None:
            print("You must specify a source with --source when connecting to a remote Kismet server")
            sys.exit(0)

        if not self.config.source == None:
            (source, options) = kismetexternal.Datasource.parse_definition(self.config.source)

            if source == None:
                print("Could not parse the --source option; this should be a standard Kismet source definition.")
                sys.exit(0)

            self.proberet = self.datasource_probesource(source, options)

            if self.proberet == None:
                print(f"Could not configure local source {self.config.source}, check your source options and config.")
                sys.exit(0)

            if not "success" in self.proberet:
                print("Could not configure local source {}, check your source options and config.")
                if "message" in self.proberet:
                    print(self.proberet["message"])
                sys.exit(0)

            if not self.proberet["success"]:
                print("Could not configure local source {}, check your source options and config.")
                if "message" in self.proberet:
                    print(self.proberet["message"].decode('utf-8'))
                sys.exit(0)

            print("Connecting to remote server {}".format(self.config.connect))


    def run(self):
        self.kismet = kismetexternal.Datasource(self.config)

        # self.kismet.debug = True

        self.kismet.set_configsource_cb(self.datasource_configure)
        self.kismet.set_listinterfaces_cb(self.datasource_listinterfaces)
        self.kismet.set_opensource_cb(self.datasource_opensource)
        self.kismet.set_probesource_cb(self.datasource_probesource)

        r = self.kismet.start()

        if r < 0:
            return

        # If we're connecting remote, kick a newsource
        if self.proberet:
            print("Registering remote source {} {}".format(self.driverid, self.config.source))
            self.kismet.send_datasource_newsource(self.config.source, self.driverid, self.proberet['uuid'])

        self.kismet.run()

    def is_running(self):
        return self.kismet.is_running()

    def get_rtl_usb_info(self, index):
        # Allocate memory buffers
        usb_manuf = (ctypes.c_char * 256)()
        usb_product = (ctypes.c_char * 256)()
        usb_serial = (ctypes.c_char * 256)()
       
        # Call the library
        self.rtl_get_usb_strings(index, usb_manuf, usb_product, usb_serial)
       
        # If there's a smarter way to do this, patches welcome
        m = bytearray(usb_manuf)
        p = bytearray(usb_product)
        s = bytearray(usb_serial)

        # Return tuple
        return (m.partition(b'\0')[0].decode('UTF-8'), p.partition(b'\0')[0].decode('UTF-8'), s.partition(b'\0')[0].decode('UTF-8'))

    def check_rtl_bin(self):
        try:
            FNULL = open(os.devnull, 'w')
            r = subprocess.check_call([self.opts['rtlbin'], "--help"], stdout=FNULL, stderr=FNULL)
        except subprocess.CalledProcessError:
            return True
        except OSError:
            return False

        return True

    async def __rtl_433_task(self):
        """
        asyncio task for consuming output from the rtl_433 process
        """

        try:
            self.kill_433()

            cmd = [ self.opts['rtlbin'], '-F', 'json', '-M', 'level' ]

            if self.opts['device'] is not None:
                cmd.append('-d')
                cmd.append("{}".format(self.opts['device']))

            if self.opts['gain'] is not None:
                cmd.append('-g')
                cmd.append("{}".format(self.opts['gain']))

            if self.opts['channel'] is not None:
                cmd.append('-f')
                cmd.append("{}".format(self.opts['channel']))

            if self.opts['ppm'] is not None:
                cmd.append('-p')
                cmd.append("{}".format(self.opts['ppm']))

            seen_any_valid = False
            failed_once = False
            print_stderr = False

            if self.opts['debug'] is not None and self.opts['debug']:
                print_stderr = True

            self.rtl_proc = await asyncio.create_subprocess_exec(*cmd,
                    stdout=asyncio.subprocess.PIPE,
                    stderr=asyncio.subprocess.DEVNULL)

            while not self.kismet.kill_ioloop:
                line = await self.rtl_proc.stdout.readline()

                if not line:
                    break

                msg = line.decode('UTF-8').strip()

                if print_stderr:
                    print("RTL433", msg, file=sys.stderr)

                if not self.handle_json(msg):
                    raise RuntimeError('could not process response from rtl_433')

                seen_any_valid = True

            raise RuntimeError('rtl_433 process exited')

        except Exception as e:
            if not self.rtl_reconfigure:
                trackeback.print_exc(file=sys.stderr)
                print("An error occurred in rtl_433; is your USB device plugged in?  Try running rtl_433 in a terminal and confirm it can connect to your device.  Make sure that no other programs are using this rtlsdr radio.", file=sys.stderr)

                self.kismet.send_datasource_error_report(message="Error handling data from rtl_433: {}".format(e))

        finally:
            self.kill_433()
            if not self.rtl_configure:
                self.kismet.spindown()

    def kill_433(self):
        try:
            if not self.rtl_proc == None:
                self.rtl_proc.kill()
        except Exception as e:
            pass

    def run_rtl433(self):
        self.kismet.add_exit_callback(self.kill_433)
        self.kismet.add_task(self.__rtl_433_task)

    # Implement the listinterfaces callback for the datasource api;
    def datasource_listinterfaces(self, seqno):
        interfaces = []

        if not self.check_rtl_bin():
            self.kismet.send_datasource_interfaces_report(seqno, interfaces)
            return

        if self.rtllib != None:
            for i in range(0, self.rtl_get_device_count()):
                (manuf, product, serial) = self.get_rtl_usb_info(i)

                dev_index = i

                # Block out empty serial numbers, and serial numbers like '1'; it might be total garbage,
                # if so, just use the index
                if len(serial) > 3:
                    dev_index = serial

                intf = kismetexternal.datasource_pb2.SubInterface()
                intf.interface = f"rtl433-{dev_index}"
                intf.capinterface = f"rtl-{dev_index}"
                intf.flags = ""
                intf.hardware = self.rtl_get_device_name(i)
                interfaces.append(intf)

        self.kismet.send_datasource_interfaces_report(seqno, interfaces)

    def __get_rtlsdr_uuid(self, intnum):
        # Get the USB info
        (manuf, product, serial) = self.get_rtl_usb_info(intnum)

        # Hash the slot, manuf, product, and serial, to get a unique ID for the UUID
        devicehash = kismetexternal.Datasource.adler32("{}{}{}{}".format(intnum, manuf, product, serial))
        devicehex = "0000{:02X}".format(devicehash)

        return kismetexternal.Datasource.make_uuid("kismet_cap_sdr_rtl433", devicehex)

    # Implement the probesource callback for the datasource api
    def datasource_probesource(self, source, options):
        ret = {}

        # Does the source look like 'rtl433-XYZ'?
        if not source[:7] == "rtl433-":
            return None

        # Do we have librtl?
        if not self.have_librtl:
            return None

        if not self.check_rtl_bin():
            return None

        # Device selector could be integer position, or it could be a serial number
        devselector = source[7:]
        intnum = -1

        # Try it as a serial number; try this first to deal with serial numbers like 00000000001
        intnum = self.rtl_get_index_by_serial(devselector.encode('utf-8'))

        # Try to find the device as an index
        if intnum < 0:
            try:
                intnum = int(devselector)

                # Abort if we're not w/in the range
                if intnum >= self.rtl_get_device_count():
                    raise ValueError("n/a")

            except ValueError:
                intnum = -1
            except:
                # Otherwise something failed in querying the hw at a deeper level
                return None

        # We've failed as both a serial and as an index, give up
        if intnum < 0:
            return None

        ret['hardware'] = self.rtl_get_device_name(intnum)

        if ('uuid' in options):
            ret['uuid'] = options['uuid']
        else:
            ret['uuid'] = self.__get_rtlsdr_uuid(intnum)

        ret['channel'] = self.opts['channel']
        ret['channels'] = [self.opts['channel']]
        ret['success'] = True
        return ret

    def datasource_opensource(self, source, options):
        ret = {}

        # Does the source look like 'rtl433-XYZ'?
        if not source[:7] == "rtl433-" and not source[:4] == "rtl-":
            ret['success'] = False
            ret['message'] = "Could not parse which rtlsdr device to use"
            return ret

        intnum = -1

        if not self.have_librtl:
            ret['success'] = False
            ret['message'] = "could not find librtlsdr, unable to configure rtlsdr interfaces"
            return ret

        # Device selector could be integer position, or it could be a serial number
        devselector = source[7:]
        found_interface = False
        intnum = -1

        # Try to find the device as an index
        try:
            intnum = int(devselector)

            # Abort if we're not w/in the range
            if intnum >= self.rtl_get_device_count():
                raise ValueError("n/a")

            # Otherwise we've found a device
            found_interface = True

        # Do nothing with exceptions; they just mean we need to look at it like a 
        # serial number
        except ValueError:
            pass

        # Try it as a serial number
        if not found_interface:
            intnum = self.rtl_get_index_by_serial(devselector.encode('utf-8'))

        # We've failed as both a serial and as an index, give up
        if intnum < 0:
            ret['success'] = False
            ret['message'] = "Could not find rtl-sdr device {}".format(devselector)
            return ret

        if 'channel' in options:
            self.opts['channel'] = options['channel']

        if 'gain' in options:
            self.opts['gain'] = options['gain']

        if 'ppm_error' in options:
            self.opts['ppm'] = options['ppm_error']

        if 'debug' in options:
            if options['debug'] == 'True' or options['debug'] == 'true':
                self.opts['debug'] = True

        ret['hardware'] = self.rtl_get_device_name(intnum)
        if ('uuid' in options):
            ret['uuid'] = options['uuid']
        else:
            ret['uuid'] = self.__get_rtlsdr_uuid(intnum)

        ret['capture_interface'] = f"rtl-{devselector}"

        self.opts['device'] = intnum

        ret['success'] = True

        self.run_rtl433()

        return ret

    def datasource_configure(self, seqno, config):
        return {"success": True}

    def handle_json(self, injson):
        try:
            j = json.loads(injson)
            r = json.dumps(j)

            report = kismetexternal.datasource_pb2.SubJson()

            dt = datetime.now()
            report.time_sec = int(time.mktime(dt.timetuple()))
            report.time_usec = int(dt.microsecond)

            report.type = "RTL433"
            report.json = r

            # print("python sending json report", r);

            self.kismet.send_datasource_data_report(full_json=report)
        except ValueError as e:
            self.kismet.send_datasource_error_report(message = "Could not parse JSON output of rtl_433")
            return False
        except Exception as e:
            self.kismet.send_datasource_error_report(message = "Could not process output of rtl_433")
            return False

        return True



import ctypes

__version__ = "2020.10.01"

class RadioMissingLibrtlsdr(Exception):
    pass

class RadioOpenError(Exception):
    pass

class RadioConfigError(RadioOpenError):
    pass

class RadioOperationalError(Exception):
    pass

class RtlSdr(object):
    def __init__(self):
        try:
            found_lib = False

            try:
                self.rtllib = ctypes.CDLL("librtlsdr.so.0")
                found_lib = True
            except OSError:
                pass

            try:
                if not found_lib:
                    self.rtllib = ctypes.CDLL("librtlsdr.dylib")
                    found_lib = True
            except OSError:
                pass

            try:
                if not found_lib:
                    self.rtllib = ctypes.CDLL("librtlsdr.dll")
                    found_lib = True
            except OSError:
                pass

            if not found_lib:
                raise OSError("could not find librtlsdr")

            self.rtl_get_device_count = self.rtllib.rtlsdr_get_device_count

            self.rtl_get_device_name = self.rtllib.rtlsdr_get_device_name
            self.rtl_get_device_name.argtypes = [ctypes.c_int]
            self.rtl_get_device_name.restype = ctypes.c_char_p

            self.rtl_get_usb_strings = self.rtllib.rtlsdr_get_device_usb_strings
            self.rtl_get_usb_strings.argtypes = [ctypes.c_int, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p]

            self.rtl_get_index_by_serial = self.rtllib.rtlsdr_get_index_by_serial
            self.rtl_get_index_by_serial.argtypes = [ctypes.c_char_p]
            self.rtl_get_index_by_serial.restype = ctypes.c_int

            self.rtl_open = self.rtllib.rtlsdr_open
            self.rtl_open.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.c_uint]
            self.rtl_open.restype = ctypes.c_int

            self.rtl_set_tuner_gain_mode = self.rtllib.rtlsdr_set_tuner_gain_mode
            self.rtl_set_tuner_gain_mode.argtypes = [ctypes.c_void_p, ctypes.c_int]
            self.rtl_set_tuner_gain_mode.restype = ctypes.c_int

            self.rtl_set_tuner_gain = self.rtllib.rtlsdr_set_tuner_gain
            self.rtl_set_tuner_gain.argtypes = [ctypes.c_void_p, ctypes.c_int]
            self.rtl_set_tuner_gain.restype = ctypes.c_int

            self.rtl_set_agc_mode = self.rtllib.rtlsdr_set_agc_mode
            self.rtl_set_agc_mode.argtypes = [ctypes.c_void_p, ctypes.c_int]
            self.rtl_set_agc_mode.restype = ctypes.c_int

            self.rtl_set_freq_correction = self.rtllib.rtlsdr_set_freq_correction
            self.rtl_set_freq_correction.argtypes = [ctypes.c_void_p, ctypes.c_int]
            self.rtl_set_freq_correction.restype = ctypes.c_int

            self.rtl_set_center_freq = self.rtllib.rtlsdr_set_center_freq
            self.rtl_set_center_freq.argtypes = [ctypes.c_void_p, ctypes.c_uint]
            self.rtl_set_center_freq.restype = ctypes.c_int

            self.rtl_get_center_freq = self.rtllib.rtlsdr_get_center_freq
            self.rtl_get_center_freq.argtypes = [ctypes.c_void_p]
            self.rtl_get_center_freq.restype = ctypes.c_int

            self.rtl_set_sample_rate = self.rtllib.rtlsdr_set_sample_rate
            self.rtl_set_sample_rate.argtypes = [ctypes.c_void_p, ctypes.c_int]
            self.rtl_set_sample_rate.restype = ctypes.c_int

            self.rtl_reset_buffer = self.rtllib.rtlsdr_reset_buffer
            self.rtl_reset_buffer.argtypes = [ctypes.c_void_p]
            self.rtl_reset_buffer.restype = ctypes.c_int

            self.rtl_read_async_cb_t = ctypes.CFUNCTYPE(None, ctypes.POINTER(ctypes.c_ubyte), ctypes.c_uint, ctypes.c_void_p)
            # self.rtl_read_async_cb = self.rtl_read_async_cb_t(self.rtl_data_cb)

            self.rtl_read_async = self.rtllib.rtlsdr_read_async
            self.rtl_read_async.argtypes = [ctypes.c_void_p, self.rtl_read_async_cb_t, ctypes.c_void_p, ctypes.c_uint, ctypes.c_uint]
            self.rtl_read_async.restype = ctypes.c_int

            self.rtl_cancel_async = self.rtllib.rtlsdr_cancel_async
            self.rtl_cancel_async.argtypes = [ctypes.c_void_p]
            self.rtl_cancel_async.restype = None

            self.rtl_set_freq_correction = self.rtllib.rtlsdr_set_freq_correction
            self.rtl_set_freq_correction.argtypes = [ctypes.c_void_p, ctypes.c_int]
            self.rtl_set_freq_correction.restype = ctypes.c_int

            try:
                self.rtl_set_bias_tee = self.rtllib.rtlsdr_set_bias_tee
                self.rtl_set_bias_tee.argtypes = [ctypes.c_void_p, ctypes.c_int]
                self.rtl_set_bias_tee.restype = ctypes.c_int
            except AttributeError:
                self.rtl_set_bias_tee = self.no_set_bias_tee

        except OSError:
            raise RadioMissingLibrtlsdr("missing librtlsdr, or it is not in your library path.")

    def no_set_bias_tee(self, foo, bar):
        print("This version of librtlsdr does not support enabling bias-tee, please upgrade your librtlsdr")
        return 0

    def get_device_count(self):
        return self.rtl_get_device_count()

    def get_rtl_usb_info(self, index):
        # Allocate memory buffers
        usb_manuf = (ctypes.c_char * 256)()
        usb_product = (ctypes.c_char * 256)()
        usb_serial = (ctypes.c_char * 256)()
       
        # Call the library
        self.rtl_get_usb_strings(index, usb_manuf, usb_product, usb_serial)
       
        # If there's a smarter way to do this, patches welcome
        m = bytearray(usb_manuf)
        p = bytearray(usb_product)
        s = bytearray(usb_serial)

        # Return tuple
        return (m.partition(b'\0')[0].decode('UTF-8'), p.partition(b'\0')[0].decode('UTF-8'), s.partition(b'\0')[0].decode('UTF-8'))

    def cancel(self):
        self.rtl_cancel_async(self.rtlradio)

    def open_radio(self, rnum, frequency, rate, gain = -1, autogain = False, ppm = 0, biastee = -1):
        self.rtlradio = ctypes.c_void_p(0)

        r = self.rtl_open(ctypes.byref(self.rtlradio), rnum)
        if not r == 0:
            raise RadioOpenError("Could not open radio")

        if not gain < 0:
            r = self.rtl_set_tuner_gain_mode(self.rtlradio, 1)
            if not r == 0:
                raise RadioConfigError("Could not set tuner gain mode")

            r = self.rtl_set_tuner_gain(self.rtlradio, gain)
            if not r == 0:
                raise RadioConfigError("Could not set gain {}".format(gain))
        elif autogain:
            r = self.rtl_set_tuner_gain_mode(self.rtlradio, 0)
            if not r == 0:
                raise RadioConfigError("Could not set tuner gain mode")
            
            r = self.rtl_set_agc_mode(self.rtlradio, 1)
            if not r == 0:
                raise RadioConfigError("Could not set agc mode")

        r = self.rtl_set_center_freq(self.rtlradio, frequency)
        if not r == 0:
            raise RadioConfigError("Could not set frequency")

        r = self.rtl_set_sample_rate(self.rtlradio, rate)
        if not r == 0:
            raise RadioConfigError("Could not set rate")

        if not ppm == 0:
            r = self.rtl_set_freq_correction(self.rtlradio, ppm)
            if not r == 0:
                raise RadioConfigError("Could not set PPM correction")

        if biastee > 0:
            r = self.rtl_set_bias_tee(self.rtlradio, 1)
            if not r == 0:
                raise RadioConfigError("Could not set bias-tee")

        r = self.rtl_reset_buffer(self.rtlradio)
        if not r == 0:
            raise RadioConfigError("Could not reset radio buffer")

    def read_samples(self, callback, nbufs, bufsz):
        """
        Read samples, calling the callback 'callback' which must be a
        rtl_read_async_cb_t function.
        """
        
        rtl_read_async_cb = self.rtl_read_async_cb_t(callback)
        r = self.rtl_read_async(self.rtlradio, rtl_read_async_cb, None, nbufs, bufsz)

        if not r == 0:
            raise RadioOpenError(f"Error reading from rtlsdr: {r}")

    def get_rtl_usb_info(self, index):
        # Allocate memory buffers
        usb_manuf = (ctypes.c_char * 256)()
        usb_product = (ctypes.c_char * 256)()
        usb_serial = (ctypes.c_char * 256)()
       
        # Call the library
        self.rtl_get_usb_strings(index, usb_manuf, usb_product, usb_serial)
       
        # If there's a smarter way to do this, patches welcome
        m = bytearray(usb_manuf)
        p = bytearray(usb_product)
        s = bytearray(usb_serial)

        # Return tuple
        return (m.partition(b'\0')[0].decode('UTF-8'), p.partition(b'\0')[0].decode('UTF-8'), s.partition(b'\0')[0].decode('UTF-8'))




"""
rtladsb Kismet data source

Sources are generated as rtladsb-XYZ when multiple rtl radios are detected.

Accepts standard options:
    channel=freq      (in raw hz)

Additionally accepts:
    ppm     error offset 
    gain    fixed gain 

"""

from __future__ import print_function

import asyncio
import argparse
import ctypes
from datetime import datetime
import json
import math

try:
    import numpy as np
except ImportError as e:
    raise ImportError("KismetRtladsb requires numpy!")

import os
import pkgutil
import subprocess
import sys
import threading
import time
import uuid

from . import rtlsdr
from . import kismetexternal

class KismetRtladsb(object):
    def __init__(self):
        self.opts = {}

        self.opts['channel'] = "1090.000MHz"
        self.opts['gain'] = -1
        self.opts['ppm'] = 0
        self.opts['device'] = None
        self.opts['debug'] = None
        self.opts['biastee'] = -1

        self.kismet = None

        self.frequency = 1090000000
        self.rate = 2000000

        self.preamble_len = 16

        self.long_frame = 112
        self.short_frame = 56
        self.long_frame_b = int(self.long_frame / 8)
        self.short_frame_b = int(self.short_frame / 8)

        self.allowed_errors = 5
        self.usb_buf_sz = 16 * 16384

        self.square_lut = np.zeros(256)
        for i in range(0, 256):
            self.square_lut[i] = abs(127 - i)
            self.square_lut[i] *= self.square_lut[i]

        # We're usually not remote
        self.proberet = None

        # Do we have librtl?
        self.have_librtl = False

        # Asyncio queue we use to post events from the SDR 
        # callback
        self.message_queue = asyncio.Queue()

        self.driverid = "rtladsb"

        try:
            self.rtlsdr = rtlsdr.RtlSdr()
            self.have_librtl = True
        except rtlsdr.RadioMissingLibrtlsdr:
            self.have_librtl = False

        parser = argparse.ArgumentParser(description='RTLadsb to Kismet bridge - Creates a rtladsb data source on a Kismet server and passes JSON-based records from the rtladsb binary')
      
        # Append the default args
        parser = kismetexternal.ExternalInterface.common_getopt(parser)
        
        self.config = parser.parse_args()

        if (self.config.infd == None or self.config.outfd == None) and self.config.connect == None:
            print("This tool is launched via Kismet IPC for local capture; see --help and the Kismet documentation to configure it for remote capture.")
            sys.exit(0)

        if not self.config.connect == None and self.config.source == None:
            print("You must specify a source with --source when connecting to a remote Kismet server")
            sys.exit(0)

        if not self.config.source == None:
            (source, options) = kismetexternal.Datasource.parse_definition(self.config.source)

            if source == None:
                print("Could not parse the --source option; this should be a standard Kismet source definition.")
                sys.exit(0)

            self.proberet = self.datasource_probesource(source, options)

            if self.proberet == None:
                print(f"Could not configure local source {self.config.source}, check your source options and config.")
                sys.exit(0)

            if not "success" in self.proberet:
                print("Could not configure local source {}, check your source options and config.")
                if "message" in self.proberet:
                    print(self.proberet["message"])
                sys.exit(0)

            if not self.proberet["success"]:
                print("Could not configure local source {}, check your source options and config.")
                if "message" in self.proberet:
                    print(self.proberet["message"])
                sys.exit(0)

            print("Connecting to remote server {}".format(self.config.connect))

    def run(self):
        self.kismet = kismetexternal.Datasource(self.config)

        self.kismet.set_configsource_cb(self.datasource_configure)
        self.kismet.set_listinterfaces_cb(self.datasource_listinterfaces)
        self.kismet.set_opensource_cb(self.datasource_opensource)
        self.kismet.set_probesource_cb(self.datasource_probesource)

        t = self.kismet.start()

        # If we're connecting remote, kick a newsource
        if self.proberet:
            print("Registering remote source {} {}".format(self.driverid, self.config.source))
            self.kismet.send_datasource_newsource(self.config.source, self.driverid, self.proberet['uuid'])

        self.kismet.run()

    def is_running(self):
        return self.kismet.is_running()

    def __get_rtlsdr_uuid(self, intnum):
        # Get the USB info
        (manuf, product, serial) = self.rtlsdr.get_rtl_usb_info(intnum)

        # Hash the slot, manuf, product, and serial, to get a unique ID for the UUID
        devicehash = kismetexternal.Datasource.adler32("{}{}{}{}".format(intnum, manuf, product, serial))
        devicehex = "0000{:02X}".format(devicehash)

        return kismetexternal.Datasource.make_uuid("kismet_cap_sdr_rtladsb", devicehex)

    async def __rtl_adsb_task(self):
        """
        asyncio task that consumes the output from the radio
        """
        print_stderr = False

        if self.opts['debug'] is not None and self.opts['debug']:
            print_stderr = True

        try:
            while not self.kismet.kill_ioloop:
                msg = await self.message_queue.get()

                if not msg:
                    break

                output = {}

                msgtype = self.adsb_msg_get_type(msg)
                msgbits = self.adsb_len_by_type(msgtype)
                msgcrc = self.adsb_msg_get_crc(msg, msgbits)
                msgcrc2 = self.adsb_crc(msg, msgbits)

                output['adsb_msg_type'] = msgtype
                output['adsb_raw_msg'] = msg.hex()
                output['crc_valid'] = False

                if msgcrc != msgcrc2:
                    if msgtype == 11 or msgtype == 17:
                        msg2 = self.adsb_msg_fix_single_bit(msg, msgbits)
     
                        if msg2 != None:
                            msg = msg2
                            output['crc_valid'] = True
                            output['crc_recovered'] = 1
                else:
                    output['crc_valid'] = True

                # Process valid messages
                if output['crc_valid']:
                    output['adsb_msg'] = msg.hex()

                    msgicao = self.adsb_msg_get_icao(msg).hex()

                    output['icao'] = msgicao

                    if msgtype == 17:
                        msgme, msgsubme = self.adsb_msg_get_me_subme(msg)

                        if msgme >= 1 and msgme <= 4:
                            msgflight = self.adsb_msg_get_flight(msg)
                            output['callsign'] = msgflight

                        elif msgme >= 9 and msgme <= 18:
                            msgalt = self.adsb_msg_get_ac12_altitude(msg)
                            output['altitude'] = msgalt

                            msgpair, msglat, msglon = self.adsb_msg_get_airborne_position(msg)
                            output['coordpair_even'] = msgpair
                            output['raw_lat'] = msglat
                            output['raw_lon'] = msglon

                        elif msgme == 19 and (msgsubme >= 1 and msgsubme <= 4):
                            if msgsubme == 1 or msgsubme == 2:
                                msgvelocity = self.adsb_msg_get_airborne_velocity(msg)
                                msgheading = self.adsb_msg_get_airborne_heading(msg)

                                output['speed'] = msgvelocity
                                output['heading'] = msgheading
                            elif msgsubme == 3 or msgsubme == 4:
                                msgheadvalid, msgheading = self.adsb_msg_get_airborne_heading(msg)
                                if msgheadvalid:
                                    out['heading'] = heading
            
                    elif msgtype == 0 or msgtype == 4 or msgtype == 16 or msgtype == 20:
                        msgalt = self.adsb_msg_get_ac13_altitude(msg)
                        output['altitude'] = msgalt

                if print_stderr:
                    print(output, file=sys.stderr)

                l = json.dumps(output)

                if not self.handle_json(l):
                    raise RuntimeError('could not process response from rtladsb')
        except Exception as e:
            traceback.print_exc(file=sys.stderr)
            print("An error occurred reading from the rtlsdr; is your USB device plugged in?  Make sure that no other programs are using this rtlsdr radio.", file=sys.stderr);

            self.kismet.send_datasource_error_report(message = "Error handling ADSB: {}".format(e))

        finally:
            self.kill_adsb()
            self.kismet.spindown()
            return

    def kill_adsb(self):
        try:
            self.rtlsdr.rtl_cancel_async(self.rtlradio)
        except:
            pass

    def run_rtladsb(self):
        self.kismet.add_exit_callback(self.kill_adsb)
        self.kismet.add_task(self.__rtl_adsb_task)
        # self.open_radio(self.opts['device'])

    # Implement the listinterfaces callback for the datasource api;
    def datasource_listinterfaces(self, seqno):
        interfaces = []

        if self.rtlsdr != None:
            for i in range(0, self.rtlsdr.get_device_count()):
                (manuf, product, serial) = self.rtlsdr.get_rtl_usb_info(i)

                dev_index = i

                # Block out empty serial numbers, and serial numbers like '1'; it might be total garbage,
                # if so, just use the index
                if len(serial) > 3:
                    dev_index = serial

                intf = kismetexternal.datasource_pb2.SubInterface()
                intf.interface = f"rtladsb-{dev_index}"
                intf.capinterface = f"rtl-{dev_index}"
                intf.flags = ""
                intf.hardware = self.rtlsdr.rtl_get_device_name(i)
                interfaces.append(intf)

        self.kismet.send_datasource_interfaces_report(seqno, interfaces)

    # Implement the probesource callback for the datasource api
    def datasource_probesource(self, source, options):
        ret = {}

        # Does the source look like 'rtladsb-XYZ'?
        if not source[:8] == "rtladsb-":
            return None

        # Do we have librtl?
        if not self.have_librtl:
            return None

        # Device selector could be integer position, or it could be a serial number
        devselector = source[8:]

        intnum = -1

        # Try it as a serial number; try this first to deal with serial numbers like 00000000001
        intnum = self.rtlsdr.rtl_get_index_by_serial(devselector.encode('utf-8'))

        # Try to find the device as an index
        if intnum < 0:
            try:
                intnum = int(devselector)

                # Abort if we're not w/in the range
                if intnum >= self.rtlsdr.rtl_get_device_count():
                    raise ValueError("n/a")

            except ValueError:
                intnum = -1
            except:
                # Otherwise something failed in querying the hw at a deeper level
                return None

        # We've failed as both a serial and as an index, give up
        if intnum < 0:
            return None

        ret['hardware'] = self.rtlsdr.rtl_get_device_name(intnum)
        if ('uuid' in options):
            ret['uuid'] = options['uuid']
        else:
            ret['uuid'] = self.__get_rtlsdr_uuid(intnum)

        ret['channel'] = self.opts['channel']
        ret['channels'] = [self.opts['channel']]
        ret['success'] = True
        return ret

    def datasource_opensource(self, source, options):
        ret = {}

        # Does the source look like 'rtladsb-XYZ'?
        if not source[:8] == "rtladsb-" and not source[:4] == "rtl-":
            ret["success"] = False
            ret["message"] = "Could not parse which rtlsdr device to use"
            return ret

        intnum = -1

        if not self.have_librtl:
            ret["success"] = False
            ret["message"] = "could not find librtlsdr, unable to configure rtlsdr interfaces"
            return ret

        # Device selector could be integer position, or it could be a serial number
        devselector = source[8:]
        found_interface = False
        intnum = -1

        # Try to find the device as an index
        try:
            intnum = int(devselector)

            # Abort if we're not w/in the range
            if intnum >= self.rtlsdr.rtl_get_device_count():
                raise ValueError("n/a")

            # Otherwise we've found a device
            found_interface = True

        except ValueError:
            # A value error means we just need to look at it as a device num
            pass
        except:
            # Otherwise something failed in querying the hw at a deeper level
            ret["success"] = False
            ret["message"] = "could not find rtlsdr device"
            return ret

        # Try it as a serial number
        if not found_interface:
            intnum = self.rtlsdr.rtl_get_index_by_serial(devselector.encode('utf-8'))

        # We've failed as both a serial and as an index, give up
        if intnum < 0:
            ret['success'] = False
            ret['message'] = "Could not find rtl-sdr device {}".format(devselector)
            return ret

        if 'debug' in options:
            if options['debug'] == 'True' or options['debug'] == 'true':
                self.opts['debug'] = True

        if 'channel' in options:
            self.opts['channel'] = options['channel']

        if 'ppm' in options:
            self.opts['ppm'] = options['ppm']

        if 'biastee' in options:
            if options['biastee'] == 'True' or options['biastee'] == 'true':
                self.opts['biastee'] = True

        if 'gain' in options:
            self.opts['gain'] = options['gain']

        ret['hardware'] = self.rtlsdr.rtl_get_device_name(intnum)
        if ('uuid' in options):
            ret['uuid'] = options['uuid']
        else:
            ret['uuid'] = self.__get_rtlsdr_uuid(intnum)

        ret['capture_interface'] = f"rtl-{devselector}"

        self.opts['device'] = intnum

        (ret['success'], ret['message']) = self.open_radio(self.opts['device'])

        if not ret['success']:
            return ret

        self.run_rtladsb()

        return ret

    def datasource_configure(self, seqno, config):
        #print(config)

        return {"success": True}

    def handle_json(self, injson):
        try:
            j = json.loads(injson)
            r = json.dumps(j)

            report = kismetexternal.datasource_pb2.SubJson()

            dt = datetime.now()
            report.time_sec = int(time.mktime(dt.timetuple()))
            report.time_usec = int(dt.microsecond)

            report.type = "RTLadsb"
            report.json = r

            # print("python sending json report", r);

            self.kismet.send_datasource_data_report(full_json=report)
        except ValueError as e:
            self.kismet.send_datasource_error_report(message = "Could handle JSON output")
            return False
        except Exception as e:
            self.kismet.send_datasource_error_report(message = "Could not handle output")
            return False

        return True

    # Raw ADSB decode of the IQ data and manchester encoded data,
    # turning it into packets.  Referenced from the rtl_adsb implementation
    # but rewritten for numpy and other python semantics
    def _iq_magnitude(self, buf, buflen):
        """
        Convert IQ to magnitude
        """

        nb = np.ctypeslib.as_array(buf, shape=(buflen,)).astype(np.uint8)

        self.magnitude_buf = np.add(self.square_lut[nb[::2]], self.square_lut[nb[1::2]])
        # self.magnitude_buf = ((np.abs(127 - nb[::2]) ** 2) + (np.abs(127 - nb[1::2]) ** 2))

    def _single_manchester(self, a, b, c, d):
        bit_p = a > b
        bit = c > d

        if bit and bit_p and c > b:
            return 1
        if bit and not bit_p and d < b:
            return 1
        if not bit and bit_p and d > b:
            return 0
        if not bit and not bit_p and c < b:
            return 0

        return None

    def _adsb_preamble(self, buf, i):
        low = 0
        high = 65535

        for i2 in range(0, self.preamble_len):
            if i2 == 0 or i2 == 2 or i2 == 7 or i2 == 9:
                high = buf[i + i2]
            else:
                low = buf[i + i2]

            if high <= low:
                return 0

        return 1

    def _adsb_np_preamble(self, buf, i):
        preamble = np.array([1, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0]) - 0.25 
        corr = np.correlate(buf[i:], preamble)
        return np.argmax(corr) + i

    def _manchester(self):
        a = 0
        b = 0

        i = 0
        while True:
            if i >= len(self.magnitude_buf) - 1:
                break

            p = self._adsb_np_preamble(self.magnitude_buf, i)

            if p + self.preamble_len >= len(self.magnitude_buf):
                break

            a = self.magnitude_buf[p]
            b = self.magnitude_buf[p + 1]
            i = p + self.preamble_len

            # # Look for a preamble
            # for ix in range(i, (len(self.magnitude_buf) - self.preamble_len)):
            #     if not self._adsb_preamble(self.magnitude_buf, ix):
            #         i = ix
            #         continue

            #     a = self.magnitude_buf[ix]
            #     b = self.magnitude_buf[ix + 1]
            #     i = ix + self.preamble_len
            #     break

            errors = 0
            m_i = 0

            message_buf = bytearray(b'\xFF' * self.long_frame)

            # Read the message until we get errors
            for ix in range(i, len(self.magnitude_buf) - 1, 2):
                i = ix + 1

                bit = self._single_manchester(a, b, self.magnitude_buf[ix], self.magnitude_buf[ix + 1])

                a = self.magnitude_buf[ix]
                b = self.magnitude_buf[ix + 1]

                if bit == None:
                    errors += 1

                    if errors > self.allowed_errors:
                        message_buf = message_buf[:m_i]
                        break
                    else:
                        if a > b:
                            bit = 1
                        else:
                            bit = 0
                        a = 0
                        b = 65535

                message_buf[m_i] = bit
                m_i = m_i + 1

                if m_i >= self.long_frame:
                    break

            if m_i < self.short_frame:
                continue
            self._adsb_message(message_buf)

    def _adsb_message(self, message_buf):
        msg_hdr = np.packbits(message_buf[0])
        if msg_hdr == 0:
            return

        adsb_frame = None
        frame_len = self.long_frame

        if msg_hdr & 0x80:
            frame_len = self.long_frame
        else:
            frame_len = self.short_frame

        if len(message_buf) < frame_len:
            return

        adsb_frame = bytearray(np.packbits(message_buf[:frame_len]).tobytes())
        
        if frame_len > self.short_frame:
            # print("*{};".format(adsb_frame.hex()))
            self.kismet.add_task(self.message_queue.put, [adsb_frame])

    def rtl_data_cb(self, buf, buflen, ctx):
        self._iq_magnitude(buf, buflen)
        self._manchester()
        return 

    def __async_radio_thread(self):
        # This function blocks forever until cancelled
        try:
            self.rtlsdr.read_samples(self.rtl_data_cb, 12, self.usb_buf_sz)
        except rtlsdr.RadioOperationalError as e:
            if not self.kismet.inSpindown():
                self.kismet.send_datasource_error_report(message = f"Error reading from RTLSDR: {e}")
        except Exception as e:
            self.kismet.send_datasource_error_report(message = f"Error reading from RTLSDR: {e}")

        # Always make sure we die
        self.kill_adsb()
        self.kismet.spindown()

    def open_radio(self, rnum):
        try:
            self.rtlsdr.open_radio(rnum, self.frequency, self.rate, gain=self.opts['gain'], autogain=True, ppm=self.opts['ppm'], biastee=self.opts['biastee'])
        except Exception as e:
            return [False, f"Error opening RTLSDR device: {e.args[0]}"]

        self.rtl_thread = threading.Thread(target=self.__async_radio_thread)
        self.rtl_thread.daemon = True
        self.rtl_thread.start()

        return [True, ""]

    # ADSB parsing functions ported from the dump1090 C implementation
    def adsb_crc(self, data, bits):
        """
        Compute the checksum a message *should* have
    
        data - bytearray 
        bits - number of bits in message
    
        return - 24-bit checksum
        """
        modes_checksum_table = [
                0x3935ea, 0x1c9af5, 0xf1b77e, 0x78dbbf, 0xc397db, 0x9e31e9, 
                0xb0e2f0, 0x587178, 0x2c38bc, 0x161c5e, 0x0b0e2f, 0xfa7d13, 
                0x82c48d, 0xbe9842, 0x5f4c21, 0xd05c14, 0x682e0a, 0x341705, 
                0xe5f186, 0x72f8c3, 0xc68665, 0x9cb936, 0x4e5c9b, 0xd8d449,
                0x939020, 0x49c810, 0x24e408, 0x127204, 0x093902, 0x049c81, 
                0xfdb444, 0x7eda22, 0x3f6d11, 0xe04c8c, 0x702646, 0x381323, 
                0xe3f395, 0x8e03ce, 0x4701e7, 0xdc7af7, 0x91c77f, 0xb719bb, 
                0xa476d9, 0xadc168, 0x56e0b4, 0x2b705a, 0x15b82d, 0xf52612,
                0x7a9309, 0xc2b380, 0x6159c0, 0x30ace0, 0x185670, 0x0c2b38, 
                0x06159c, 0x030ace, 0x018567, 0xff38b7, 0x80665f, 0xbfc92b, 
                0xa01e91, 0xaff54c, 0x57faa6, 0x2bfd53, 0xea04ad, 0x8af852, 
                0x457c29, 0xdd4410, 0x6ea208, 0x375104, 0x1ba882, 0x0dd441,
                0xf91024, 0x7c8812, 0x3e4409, 0xe0d800, 0x706c00, 0x383600, 
                0x1c1b00, 0x0e0d80, 0x0706c0, 0x038360, 0x01c1b0, 0x00e0d8, 
                0x00706c, 0x003836, 0x001c1b, 0xfff409, 0x000000, 0x000000, 
                0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000,
                0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 
                0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 
                0x000000, 0x000000, 0x000000, 0x000000 ]
    
        crc = 0
        offset = 0
    
        if bits != 112:
            offset = 112 - 56
    
        for j in range(0, bits):
            byte = int(j / 8)
            bit = j % 8
            bitmask = 1 << (7 - bit)
    
            if data[byte] & bitmask:
                crc ^= modes_checksum_table[j + offset]
    
        return crc & 0x00FFFFFF
    
    def adsb_len_by_type(self, type):
        """
        Get expected length of message in bits based on the type
        """
    
        if type == 16 or type == 17 or type == 19 or type == 20 or type == 21:
            return 112
    
        return 56
    
    def adsb_msg_get_crc(self, data, bits):
        """
        Extract the crc encoded in a message
    
        data - bytearray of message input
        bits - number of bits in message
    
        return - 24bit checksum as encoded in message
        """
    
        crc = (data[int(bits / 8) - 3] << 16)
        crc |= (data[int(bits / 8) - 2] << 8) 
        crc |= (data[int(bits / 8) - 1])
    
        return crc
    
    def adsb_msg_fix_single_bit(self, data, bits):
        """
        Try to fix single bit errors using the checksum.  On success
        returns modified bytearray
    
        data - bytearray of message input
        bits - length in bits
        """
    
        for j in range(0, bits):
            byte = int(j / 8)
            bitmask = 1 << (7 - (j % 8))
    
            aux = data[:]
    
            # Flip the j-th bit
            aux[byte] ^= bitmask
    
            crc1 = (aux[int(bits / 8) - 3] << 16)
            crc1 |= (aux[int(bits / 8) - 2] << 8) 
            crc1 |= (aux[int(bits / 8) - 1])
    
            crc2 = self.adsb_crc(aux, bits)
    
            if crc1 == crc2:
                # The error is fixed; return the new buffer
                return aux
    
        return None
    
    def adsb_msg_fix_double_bit(self, data, bits):
        """
        Try to fix double bit errors using the checksum, like fix_single_bit.
        This is very slow and should only be tried against DF17 messages.
        
        If successful returns the modified bytearray.
    
        data - bytearray of message input
        bits - length in bits
        """
    
        for j in range(0, bits):
            byte1 = int(j / 8)
            bitmask1 = 1 << (7 - (j % 8))
    
            # Don't check the same pairs multiple times, so i starts from j+1
            for i in range(j + 1, bits):
                byte2 = int(i / 8)
                bitmask2 = 1 << (7 - (i % 8))
    
                aux = data[:]
    
                # Flip the jth bit
                aux[byte1] ^= bitmask1
                # Flip the ith bit
                aux[byte2] ^= bitmask2
    
                crc1 = (aux[int(bits / 8) - 3] << 16)
                crc1 |= (aux[int(bits / 8) - 2] << 8) 
                crc1 |= (aux[int(bits / 8) - 1])
    
                crc2 = self.adsb_crc(aux, bits)
    
                if crc1 == crc2:
                    # The error is fixed; return the new buffer
                    return aux
    
        return None
    
    
        
    def adsb_msg_get_type(self, data):
        """
        Get message type
        """
    
        return data[0] >> 3
    
    def adsb_msg_get_icao(self, data):
        """
        Get ICAO
        """
        return data[1:4]
    
    def adsb_msg_get_fs(self, data):
        """
        Extract flight status from 4, 5, 20, 21
        """
        return data[0] & 7
    
    def adsb_msg_get_me_subme(self, data):
        """
        Extract message 17 metype and mesub type
    
        Returns:
        (type,subtype) tuple
        """
    
        return (data[4] >> 3, data[4] & 7)
    
    def adsb_msg_get_ac13_altitude(self, data):
        """
        Extract 13 bit altitude (in feet) from 0, 4, 16, 20
        """
    
        m_bit = data[3] & (1 << 6)
        q_bit = data[3] & (1 << 4)
    
        if not m_bit:
            if q_bit:
                # N is the 11 bit integer resulting in the removal of bit q and m
                n = (data[2] & 31) << 6
                n |= (data[3] & 0x80) >> 2
                n |= (data[3] & 0x20) >> 1
                n |= (data[3] & 0x15)
    
                return n * 25 - 1000
    
        return 0
    
    def adsb_msg_get_ac12_altitude(self, data):
        """
        Extract 12 bit altitude (in feet) from 17
        """
    
        q_bit = data[5] & 1
    
        if q_bit:
            # N is the 11 bit integer resulting from the removal of bit Q
            n = (data[5] >> 1) << 4
            n |= (data[6] & 0xF0) >> 4
    
            # print("Raw bytes {} {} return {}".format(data[5], data[6], n * 25 - 1000))
            return n * 25 - 1000
    
        return 0
    
    def adsb_msg_get_flight(self, data):
        """
        Extract flight name
        """
    
        ais_charset = "?ABCDEFGHIJKLMNOPQRSTUVWXYZ????? ???????????????0123456789??????"
    
        flight = ""
    
        flight += ais_charset[data[5] >> 2]
        flight += ais_charset[((data[5] & 3) << 4) | (data[6] >> 4)]
        flight += ais_charset[((data[6] & 15) << 2) | (data[7] >> 6)]
        flight += ais_charset[data[7] & 63]
        flight += ais_charset[data[8] >> 2]
        flight += ais_charset[((data[8] & 3) << 4) | (data[9] >> 4)]
        flight += ais_charset[((data[9] & 15) << 2) | (data[10] >> 6)]
        flight += ais_charset[data[10] & 63]
    
        return flight.strip()
    
    def adsb_msg_get_airborne_position(self, data):
        """
        Airborne position message from message 17
    
        Return:
        (pair, lat, lon) raw tuple of even (0) or odd (1) and raw lat/lon
        """
    
        paireven = (data[6] & (1 << 2)) == 0
    
        lat = (data[6] & 3) << 15
        lat |= data[7] << 7
        lat |= data[8] >> 1
    
        lon = (data[8] & 1) << 16
        lon |= data[9] << 8
        lon |= data[10]
    
        return (paireven, lat, lon)
    
    def adsb_msg_get_airborne_velocity(self, data):
        """
        Airborne velocity from message 17, synthesized from EW/NS velocities
        """
    
        ew_dir = (data[5] & 4) >> 2
        ew_velocity = ((data[5] & 3) << 8) | data[6]
        ns_dir = (data[7] & 0x80) >> 7
        ns_velocity = ((data[7] & 0x7f) << 3) | ((data[8] & 0xe0) >> 5)
    
        # Compute velocity from two speed components
        velocity = math.sqrt(ns_velocity * ns_velocity + ew_velocity * ew_velocity)
    
        return velocity
    
    def adsb_msg_get_airborne_heading(self, data):
        """
        Airborne heading from message 17, synthesized from EW/NS velocities
    
        Returns:
            Heading in degrees
        """
    
        ew_dir = (data[5] & 4) >> 2
        ew_velocity = ((data[5] & 3) << 8) | data[6]
        ns_dir = (data[7] & 0x80) >> 7
        ns_velocity = ((data[7] & 0x7f) << 3) | ((data[8] & 0xe0) >> 5)
    
        ewv = ew_velocity
        nsv = ns_velocity
    
        if ew_dir:
            ewv *= -1
    
        if ns_dir:
            nsv *= -1
    
        heading = math.atan2(ewv, nsv)
    
        # Convert to degrees
        heading = heading * 360 / (math.pi * 2)
    
        if heading < 0:
            heading += 360
    
        return heading
    
    def adsb_msg_get_sub3_heading(self, data):
        """
        Direct heading from msg 17 sub 3 and 4
    
        Returns:
            Heading in degrees
        """
    
        valid = data[5] & (1 << 2)
        heading = (data[5] & 3) << 5
        heading |= data[6] >> 3
        heading = heading * (360.0 / 128)
    
        return valid, heading
    
    def adsb_process_msg(self, msg):
        output = {}
     
        msgtype = self.adsb_msg_get_type(msg)
        msgbits = self.adsb_len_by_type(msgtype)
        msgcrc = self.adsb_msg_get_crc(msg, msgbits)
        msgcrc2 = self.adsb_crc(msg, msgbits)
     
        output['adsb_msg_type'] = msgtype
        output['adsb_raw_msg'] = msg.hex()
        output['crc_valid'] = False
     
        # Skip invalid CRC types; in the future, add 1bit recovery from dump1090
        if msgcrc != msgcrc2:
            if msgtype == 11 or msgtype == 17:
                msg2 = self.adsb_msg_fix_single_bit(msg, msgbits)
     
                if msg2 != None:
                    msg = msg2
                    output['crc_valid'] = True
                    output['crc_recovered'] = 1
     
        else:
            output['crc_valid'] = True
     
        # Process valid messages
        if output['crc_valid']:
            output['adsb_msg'] = msg.hex()
     
            msgicao = self.adsb_msg_get_icao(msg).hex()
     
            output['icao'] = msgicao
     
            if msgtype == 17:
                msgme, msgsubme = self.adsb_msg_get_me_subme(msg)
     
                if msgme >= 1 and msgme <= 4:
                    msgflight = self.adsb_msg_get_flight(msg)
                    output['callsign'] = msgflight
     
                elif msgme >= 9 and msgme <= 18:
                    msgalt = self.adsb_msg_get_ac12_altitude(msg)
                    output['altitude'] = msgalt
     
                    msgpair, msglat, msglon = self.adsb_msg_get_airborne_position(msg)
                    output['coordpair_even'] = msgpair
                    output['raw_lat'] = msglat
                    output['raw_lon'] = msglon
     
                elif msgme == 19 and (msgsubme >= 1 and msgsubme <= 4):
                    msgpair, msglat, msglon = self.adsb_msg_get_airborne_position(msg)
                    output['coordpair_even'] = msgpair
                    output['raw_lat'] = msglat
                    output['raw_lon'] = msglon
     
                    msgalt = self.adsb_msg_get_ac12_altitude(msg)
                    output['altitude'] = msgalt
     
                    if msgsubme == 1 or msgsubme == 2:
                        msgvelocity = self.adsb_msg_get_airborne_velocity(msg)
                        msgheading = self.adsb_msg_get_airborne_heading(msg)
     
                        output['speed'] = msgvelocity
                        output['heading'] = msgheading
                    elif msgsubme == 3 or msgsubme == 4:
                        msgheadvalid, msgheading = self.adsb_msg_get_airborne_heading(msg)
                        if msgheadvalid:
                            out['heading'] = msgheading
     
            elif msgtype == 0 or msgtype == 4 or msgtype == 16 or msgtype == 20:
                msgalt = self.adsb_msg_get_ac13_altitude(msg)
                output['altitude'] = msgalt
     
        if output['crc_valid']:
            print(output, file=sys.stderr)





# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: datasource.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import kismet_pb2 as kismet__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10\x64\x61tasource.proto\x12\x10KismetDatasource\x1a\x0ckismet.proto\"\x1f\n\x0bSubChannels\x12\x10\n\x08\x63hannels\x18\x01 \x03(\t\"\x1d\n\nSubChanset\x12\x0f\n\x07\x63hannel\x18\x01 \x02(\t\"c\n\nSubChanhop\x12\x10\n\x08\x63hannels\x18\x01 \x03(\t\x12\x0c\n\x04rate\x18\x02 \x01(\x01\x12\x0f\n\x07shuffle\x18\x03 \x01(\x08\x12\x14\n\x0cshuffle_skip\x18\x04 \x01(\r\x12\x0e\n\x06offset\x18\x05 \x01(\r\"\xc8\x01\n\x06SubGps\x12\x0b\n\x03lat\x18\x01 \x02(\x01\x12\x0b\n\x03lon\x18\x02 \x02(\x01\x12\x0b\n\x03\x61lt\x18\x03 \x02(\x01\x12\r\n\x05speed\x18\x04 \x02(\x01\x12\x0f\n\x07heading\x18\x05 \x02(\x01\x12\x11\n\tprecision\x18\x06 \x02(\x01\x12\x0b\n\x03\x66ix\x18\x07 \x02(\r\x12\x10\n\x08time_sec\x18\x08 \x02(\x04\x12\x11\n\ttime_usec\x18\t \x02(\x04\x12\x0c\n\x04type\x18\n \x02(\t\x12\x0c\n\x04name\x18\x0b \x02(\t\x12\x16\n\x0ehigh_prec_time\x18\x0c \x01(\x01\"X\n\x0cSubInterface\x12\x11\n\tinterface\x18\x01 \x02(\t\x12\r\n\x05\x66lags\x18\x02 \x02(\t\x12\x10\n\x08hardware\x18\x03 \x01(\t\x12\x14\n\x0c\x63\x61pinterface\x18\x04 \x01(\t\"k\n\tSubPacket\x12\x10\n\x08time_sec\x18\x01 \x02(\x04\x12\x11\n\ttime_usec\x18\x02 \x02(\x04\x12\x0b\n\x03\x64lt\x18\x03 \x02(\r\x12\x0c\n\x04size\x18\x04 \x02(\x04\x12\x0c\n\x04\x64\x61ta\x18\x05 \x02(\x0c\x12\x10\n\x08\x63\x61p_size\x18\x06 \x01(\x04\"J\n\x07SubJson\x12\x10\n\x08time_sec\x18\x01 \x02(\x04\x12\x11\n\ttime_usec\x18\x02 \x02(\x04\x12\x0c\n\x04type\x18\x03 \x02(\t\x12\x0c\n\x04json\x18\x04 \x02(\t\"N\n\tSubBuffer\x12\x10\n\x08time_sec\x18\x01 \x02(\x04\x12\x11\n\ttime_usec\x18\x02 \x02(\x04\x12\x0c\n\x04type\x18\x03 \x02(\t\x12\x0e\n\x06\x62uffer\x18\x04 \x02(\x0c\"\x90\x01\n\tSubSignal\x12\x12\n\nsignal_dbm\x18\x01 \x01(\x01\x12\x11\n\tnoise_dbm\x18\x02 \x01(\x01\x12\x13\n\x0bsignal_rssi\x18\x03 \x01(\x01\x12\x12\n\nnoise_rssi\x18\x04 \x01(\x01\x12\x10\n\x08\x66req_khz\x18\x05 \x01(\x01\x12\x0f\n\x07\x63hannel\x18\x06 \x01(\t\x12\x10\n\x08\x64\x61tarate\x18\x07 \x01(\x01\"\x9f\x01\n\nSubSpecset\x12\x11\n\tstart_mhz\x18\x01 \x01(\x01\x12\x0f\n\x07\x65nd_mhz\x18\x02 \x01(\x01\x12\x1a\n\x12samples_per_bucket\x18\x03 \x01(\x01\x12\x17\n\x0f\x62ucket_width_hz\x18\x04 \x01(\x01\x12\x12\n\nenable_amp\x18\x05 \x01(\x08\x12\x0e\n\x06if_amp\x18\x06 \x01(\x04\x12\x14\n\x0c\x62\x61seband_amp\x18\x07 \x01(\x04\"}\n\x0bSubSpectrum\x12\x10\n\x08time_sec\x18\x01 \x01(\x04\x12\x11\n\ttime_usec\x18\x02 \x01(\x04\x12\x11\n\tstart_mhz\x18\x03 \x01(\x01\x12\x0f\n\x07\x65nd_mhz\x18\x04 \x01(\x01\x12\x17\n\x0f\x62ucket_width_hz\x18\x05 \x01(\x01\x12\x0c\n\x04\x64\x61ta\x18\x06 \x03(\x05\",\n\nSubSuccess\x12\x0f\n\x07success\x18\x01 \x02(\x08\x12\r\n\x05seqno\x18\x02 \x02(\r\"\x11\n\x0f\x43loseDatasource\"\x99\x01\n\tConfigure\x12-\n\x07\x63hannel\x18\x01 \x01(\x0b\x32\x1c.KismetDatasource.SubChanset\x12-\n\x07hopping\x18\x02 \x01(\x0b\x32\x1c.KismetDatasource.SubChanhop\x12.\n\x08spectrum\x18\x03 \x01(\x0b\x32\x1c.KismetDatasource.SubSpecset\"\xdf\x01\n\x0f\x43onfigureReport\x12-\n\x07success\x18\x01 \x02(\x0b\x32\x1c.KismetDatasource.SubSuccess\x12-\n\x07\x63hannel\x18\x02 \x01(\x0b\x32\x1c.KismetDatasource.SubChanset\x12-\n\x07hopping\x18\x03 \x01(\x0b\x32\x1c.KismetDatasource.SubChanhop\x12.\n\x07message\x18\x04 \x01(\x0b\x32\x1d.KismetExternal.MsgbusMessage\x12\x0f\n\x07warning\x18\x05 \x01(\t\"\xed\x02\n\nDataReport\x12%\n\x03gps\x18\x01 \x01(\x0b\x32\x18.KismetDatasource.SubGps\x12.\n\x07message\x18\x02 \x01(\x0b\x32\x1d.KismetExternal.MsgbusMessage\x12+\n\x06packet\x18\x03 \x01(\x0b\x32\x1b.KismetDatasource.SubPacket\x12+\n\x06signal\x18\x04 \x01(\x0b\x32\x1b.KismetDatasource.SubSignal\x12/\n\x08spectrum\x18\x05 \x01(\x0b\x32\x1d.KismetDatasource.SubSpectrum\x12\x0f\n\x07warning\x18\x06 \x01(\t\x12\'\n\x04json\x18\x07 \x01(\x0b\x32\x19.KismetDatasource.SubJson\x12+\n\x06\x62uffer\x18\x08 \x01(\x0b\x32\x1b.KismetDatasource.SubBuffer\x12\x16\n\x0ehigh_prec_time\x18\t \x01(\x01\"l\n\x0b\x45rrorReport\x12-\n\x07success\x18\x01 \x02(\x0b\x32\x1c.KismetDatasource.SubSuccess\x12.\n\x07message\x18\x02 \x01(\x0b\x32\x1d.KismetExternal.MsgbusMessage\"\x10\n\x0eListInterfaces\"\xa5\x01\n\x10InterfacesReport\x12-\n\x07success\x18\x01 \x02(\x0b\x32\x1c.KismetDatasource.SubSuccess\x12.\n\x07message\x18\x02 \x01(\x0b\x32\x1d.KismetExternal.MsgbusMessage\x12\x32\n\ninterfaces\x18\x03 \x03(\x0b\x32\x1e.KismetDatasource.SubInterface\"A\n\tNewSource\x12\x12\n\ndefinition\x18\x01 \x02(\t\x12\x12\n\nsourcetype\x18\x02 \x02(\t\x12\x0c\n\x04uuid\x18\x03 \x02(\t\" \n\nOpenSource\x12\x12\n\ndefinition\x18\x01 \x02(\t\"\x8c\x03\n\x10OpenSourceReport\x12-\n\x07success\x18\x01 \x02(\x0b\x32\x1c.KismetDatasource.SubSuccess\x12\x0b\n\x03\x64lt\x18\x02 \x01(\r\x12\x19\n\x11\x63\x61pture_interface\x18\x03 \x01(\t\x12/\n\x08\x63hannels\x18\x04 \x01(\x0b\x32\x1d.KismetDatasource.SubChannels\x12-\n\x07\x63hannel\x18\x05 \x01(\x0b\x32\x1c.KismetDatasource.SubChanset\x12\x30\n\nhop_config\x18\x06 \x01(\x0b\x32\x1c.KismetDatasource.SubChanhop\x12\x10\n\x08hardware\x18\x07 \x01(\t\x12.\n\x07message\x18\x08 \x01(\x0b\x32\x1d.KismetExternal.MsgbusMessage\x12.\n\x08spectrum\x18\t \x01(\x0b\x32\x1c.KismetDatasource.SubSpecset\x12\x0c\n\x04uuid\x18\n \x01(\t\x12\x0f\n\x07warning\x18\x0b \x01(\t\"!\n\x0bProbeSource\x12\x12\n\ndefinition\x18\x01 \x02(\t\"\x94\x02\n\x11ProbeSourceReport\x12-\n\x07success\x18\x01 \x02(\x0b\x32\x1c.KismetDatasource.SubSuccess\x12.\n\x07message\x18\x02 \x01(\x0b\x32\x1d.KismetExternal.MsgbusMessage\x12/\n\x08\x63hannels\x18\x03 \x01(\x0b\x32\x1d.KismetDatasource.SubChannels\x12-\n\x07\x63hannel\x18\x04 \x01(\x0b\x32\x1c.KismetDatasource.SubChanset\x12.\n\x08spectrum\x18\x05 \x01(\x0b\x32\x1c.KismetDatasource.SubSpecset\x12\x10\n\x08hardware\x18\x06 \x01(\t\" \n\rWarningReport\x12\x0f\n\x07warning\x18\x01 \x02(\tB\x02H\x03')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'datasource_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'H\003'
  _SUBCHANNELS._serialized_start=52
  _SUBCHANNELS._serialized_end=83
  _SUBCHANSET._serialized_start=85
  _SUBCHANSET._serialized_end=114
  _SUBCHANHOP._serialized_start=116
  _SUBCHANHOP._serialized_end=215
  _SUBGPS._serialized_start=218
  _SUBGPS._serialized_end=418
  _SUBINTERFACE._serialized_start=420
  _SUBINTERFACE._serialized_end=508
  _SUBPACKET._serialized_start=510
  _SUBPACKET._serialized_end=617
  _SUBJSON._serialized_start=619
  _SUBJSON._serialized_end=693
  _SUBBUFFER._serialized_start=695
  _SUBBUFFER._serialized_end=773
  _SUBSIGNAL._serialized_start=776
  _SUBSIGNAL._serialized_end=920
  _SUBSPECSET._serialized_start=923
  _SUBSPECSET._serialized_end=1082
  _SUBSPECTRUM._serialized_start=1084
  _SUBSPECTRUM._serialized_end=1209
  _SUBSUCCESS._serialized_start=1211
  _SUBSUCCESS._serialized_end=1255
  _CLOSEDATASOURCE._serialized_start=1257
  _CLOSEDATASOURCE._serialized_end=1274
  _CONFIGURE._serialized_start=1277
  _CONFIGURE._serialized_end=1430
  _CONFIGUREREPORT._serialized_start=1433
  _CONFIGUREREPORT._serialized_end=1656
  _DATAREPORT._serialized_start=1659
  _DATAREPORT._serialized_end=2024
  _ERRORREPORT._serialized_start=2026
  _ERRORREPORT._serialized_end=2134
  _LISTINTERFACES._serialized_start=2136
  _LISTINTERFACES._serialized_end=2152
  _INTERFACESREPORT._serialized_start=2155
  _INTERFACESREPORT._serialized_end=2320
  _NEWSOURCE._serialized_start=2322
  _NEWSOURCE._serialized_end=2387
  _OPENSOURCE._serialized_start=2389
  _OPENSOURCE._serialized_end=2421
  _OPENSOURCEREPORT._serialized_start=2424
  _OPENSOURCEREPORT._serialized_end=2820
  _PROBESOURCE._serialized_start=2822
  _PROBESOURCE._serialized_end=2855
  _PROBESOURCEREPORT._serialized_start=2858
  _PROBESOURCEREPORT._serialized_end=3134
  _WARNINGREPORT._serialized_start=3136
  _WARNINGREPORT._serialized_end=3168
# @@protoc_insertion_point(module_scope)



"""
rtlamr Kismet data source

Sources are generated as rtlaamr-XYZ when multiple rtl radios are detected.

Accepts standard options:
    channel=freq      (in raw hz)

Additionally accepts:
    ppm     error offset 
    gain    fixed gain 

"""

from __future__ import print_function

import asyncio
import argparse
import csv
import ctypes
from datetime import datetime
import json
import math

try:
    import numpy as np
except ImportError as e:
    raise ImportError("KismetRtlamr requires numpy!")

import os
import pkgutil
import subprocess
import sys
import threading
import time
import uuid

from . import rtlsdr
from . import kismetexternal

class KismetRtlamr(object):
    def __init__(self):
        self.opts = {}

        self.opts['channel'] = "912.600MHz"
        self.opts['gain'] = -1
        self.opts['ppm'] = 0
        self.opts['device'] = None
        self.opts['debug'] = None
        self.opts['biastee'] = -1

        self.kismet = None

        self.frequency = 912600000
        self.rate = 2359000
        self.usb_buf_sz = 16 * 16384

        # At our given rate, we're 72 samples per symbol
        self.symbol_len = 72

        # Messages are 12 bytes
        self.message_len_b = 12

        # With manchester doubling the bits, get the len in samples
        self.message_len_s = 2 * self.message_len_b * self.symbol_len

        # Preamble taken before the manchester decode
        self.scm_preamble = np.array([1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, ])
        self.scm_preamble_len = len(self.scm_preamble)
        self.scm_preamble_len_s = self.scm_preamble_len * self.symbol_len

        # Generate the normalized squares for converting IQ via lookup
        self.square_lut = np.zeros(256)
        for i in range(0, 256):
            self.square_lut[i] = (127.5 - float(i)) / 127.5
            self.square_lut[i] *= self.square_lut[i]

        # BCH checksum polynomial
        self.bch_poly = 0x6F63

        # Generate the polynomial table
        self.bch_table = np.zeros(256).astype(np.uint16)
        for i in range(0, 256):
            crc = i << 8
            for n in range(0, 8):
                if not (crc & 0x8000) == 0:
                    crc = (crc << 1) ^ self.bch_poly
                else:
                    crc = crc << 1

            self.bch_table[i] = int(crc)

        # The higher the decimation the more CPU we save; we decimate AFTER
        # quantization and this seems to be consistently usable with a very high
        # dynamic range of capture
        self.decimation = 24
        self.reduced_w = int(self.symbol_len / self.decimation)
        self.reduced_preamble_l = self.reduced_w * self.scm_preamble_len

        # Expand the preamble to fit the symbol width; we only search for the 
        # first 16 bits then compare the rest
        self.search_preamble = np.repeat(self.scm_preamble[:16], self.reduced_w)

        # We're usually not remote
        self.proberet = None

        # Do we have librtl?
        self.have_librtl = False

        # Asyncio queue we use to post events from the SDR 
        # callback
        self.message_queue = asyncio.Queue()

        self.driverid = "rtlamr"

        try:
            self.rtlsdr = rtlsdr.RtlSdr()
            self.have_librtl = True
        except rtlsdr.RadioMissingLibrtlsdr:
            self.have_librtl = False

        parser = argparse.ArgumentParser(description='RTL-SDR AMR Kismet datasource')
        
        # Append the default args
        parser = kismetexternal.ExternalInterface.common_getopt(parser)
        
        self.config = parser.parse_args()

        if (self.config.infd == None or self.config.outfd == None) and self.config.connect == None:
            print("This tool is launched via Kismet IPC for local capture; see --help and the Kismet documentation to configure it for remote capture.")
            sys.exit(0)

        if not self.config.connect == None and self.config.source == None:
            print("You must specify a source with --source when connecting to a remote Kismet server")
            sys.exit(0)

        if not self.config.source == None:
            (source, options) = kismetexternal.Datasource.parse_definition(self.config.source)

            if source == None:
                print("Could not parse the --source option; this should be a standard Kismet source definition.")
                sys.exit(0)

            self.proberet = self.datasource_probesource(source, options)

            if self.proberet == None:
                print(f"Could not configure local source {config.source}, check your source options and config.")
                sys.exit(0)

            if not "success" in self.proberet:
                print("Could not configure local source {}, check your source options and config.")
                if "message" in self.proberet:
                    print(self.proberet["message"])
                sys.exit(0)

            if not self.proberet["success"]:
                print("Could not configure local source {}, check your source options and config.")
                if "message" in self.proberet:
                    print(self.proberet["message"])
                sys.exit(0)

            print("Connecting to remote server {}".format(self.config.connect))

    def run(self):
        self.kismet = kismetexternal.Datasource(self.config)

        self.kismet.set_configsource_cb(self.datasource_configure)
        self.kismet.set_listinterfaces_cb(self.datasource_listinterfaces)
        self.kismet.set_opensource_cb(self.datasource_opensource)
        self.kismet.set_probesource_cb(self.datasource_probesource)

        t = self.kismet.start()

        # If we're connecting remote, kick a newsource
        if self.proberet:
            print("Registering remote source {} {}".format(self.driverid, self.config.source))
            self.kismet.send_datasource_newsource(self.config.source, self.driverid, self.proberet['uuid'])

        self.kismet.run()

    def is_running(self):
        return self.kismet.is_running()

    def __get_rtlsdr_uuid(self, intnum):
        # Get the USB info
        (manuf, product, serial) = self.rtlsdr.get_rtl_usb_info(intnum)

        # Hash the slot, manuf, product, and serial, to get a unique ID for the UUID
        devicehash = kismetexternal.Datasource.adler32("{}{}{}{}".format(intnum, manuf, product, serial))
        devicehex = "0000{:02X}".format(devicehash)

        return kismetexternal.Datasource.make_uuid("kismet_cap_sdr_rtlamr", devicehex)

    def kill_amr(self):
        try:
            self.rtlsdr.rtl_cancel_async(self.rtlradio)
        except:
            pass

    def run_rtlamr(self):
        self.kismet.add_exit_callback(self.kill_amr)
        self.kismet.add_task(self.__rtl_amr_task)
        # self.open_radio(self.opts['device'])

    # Implement the listinterfaces callback for the datasource api;
    def datasource_listinterfaces(self, seqno):
        interfaces = []

        if self.rtlsdr != None:
            for i in range(0, self.rtlsdr.get_device_count()):
                (manuf, product, serial) = self.rtlsdr.get_rtl_usb_info(i)

                dev_index = i

                # Block out empty serial numbers, and serial numbers like '1'; it might be total garbage,
                # if so, just use the index
                if len(serial) > 3:
                    dev_index = serial

                intf = kismetexternal.datasource_pb2.SubInterface()
                intf.interface = f"rtlamr-{dev_index}"
                intf.capinterface = f"rtl-{dev_index}"
                intf.flags = ""
                intf.hardware = self.rtlsdr.rtl_get_device_name(i)
                interfaces.append(intf)

        self.kismet.send_datasource_interfaces_report(seqno, interfaces)

    # Implement the probesource callback for the datasource api
    def datasource_probesource(self, source, options):
        ret = {}

        # Does the source look like 'rtlamr-XYZ'?
        if not source[:7] == "rtlamr-":
            return None

        # Do we have librtl?
        if not self.have_librtl:
            return None

        # Device selector could be integer position, or it could be a serial number
        devselector = source[7:]
        found_interface = False

        intnum = -1

        # Try it as a serial number; try this first to deal with serial numbers like 00000000001
        intnum = self.rtlsdr.rtl_get_index_by_serial(devselector.encode('utf-8'))

        # Try to find the device as an index
        if intnum < 0:
            try:
                intnum = int(devselector)

                # Abort if we're not w/in the range
                if intnum >= self.rtlsdr.rtl_get_device_count():
                    raise ValueError("n/a")

            except ValueError:
                intnum = -1
            except:
                # Otherwise something failed in querying the hw at a deeper level
                return None

        # We've failed as both a serial and as an index, give up
        if intnum < 0:
            return None

        ret['hardware'] = self.rtlsdr.rtl_get_device_name(intnum)
        if ('uuid' in options):
            ret['uuid'] = options['uuid']
        else:
            ret['uuid'] = self.__get_rtlsdr_uuid(intnum)

        ret['channel'] = self.opts['channel']
        ret['channels'] = [self.opts['channel']]
        ret['success'] = True
        return ret

    def datasource_opensource(self, source, options):
        ret = {}

        # Does the source look like 'rtlamr-XYZ'?
        if not source[:7] == "rtlamr-" and not source[:4] == "rtl-":
            ret["success"] = False
            ret["message"] = "Could not parse which rtlsdr device to use"
            return ret

        intnum = -1

        if not self.have_librtl:
            ret["success"] = False
            ret["message"] = "could not find librtlsdr, unable to configure rtlsdr interfaces"
            return ret

        # Device selector could be integer position, or it could be a serial number
        devselector = source[7:]
        found_interface = False
        intnum = -1

        # Try to find the device as an index
        try:
            intnum = int(devselector)

            # Abort if we're not w/in the range
            if intnum >= self.rtlsdr.rtl_get_device_count():
                raise ValueError("n/a")

            # Otherwise we've found a device
            found_interface = True

        except ValueError:
            # A value error means we just need to look at it as a device num
            pass
        except:
            # Otherwise something failed in querying the hw at a deeper level
            ret["success"] = False
            ret["message"] = "could not find rtlsdr device"
            return ret

        # Try it as a serial number
        if not found_interface:
            intnum = self.rtlsdr.rtl_get_index_by_serial(devselector.encode('utf-8'))

        # We've failed as both a serial and as an index, give up
        if intnum < 0:
            ret['success'] = False
            ret['message'] = "Could not find rtl-sdr device {}".format(devselector)
            return ret

        if 'debug' in options:
            if options['debug'] == 'True' or options['debug'] == 'true':
                self.opts['debug'] = True

        if 'channel' in options:
            self.opts['channel'] = options['channel']

        if 'ppm' in options:
            self.opts['ppm'] = options['ppm']

        if 'biastee' in options:
            if options['biastee'] == 'True' or options['biastee'] == 'true':
                self.opts['biastee'] = True

        if 'gain' in options:
            self.opts['gain'] = options['gain']

        ret['hardware'] = self.rtlsdr.rtl_get_device_name(intnum)
        if ('uuid' in options):
            ret['uuid'] = options['uuid']
        else:
            ret['uuid'] = self.__get_rtlsdr_uuid(intnum)

        ret['capture_interface'] = f"rtl-{devselector}"

        self.opts['device'] = intnum

        (ret['success'], ret['message']) = self.open_radio(self.opts['device'])

        if not ret['success']:
            return ret

        self.run_rtlamr()

        return ret

    def datasource_configure(self, seqno, config):
        #print(config)

        return {"success": True}

    def handle_json(self, injson):
        try:
            j = json.loads(injson)
            r = json.dumps(j)

            report = kismetexternal.datasource_pb2.SubJson()

            dt = datetime.now()
            report.time_sec = int(time.mktime(dt.timetuple()))
            report.time_usec = int(dt.microsecond)

            report.type = "RTLamr"
            report.json = r

            # print("python sending json report", r);

            self.kismet.send_datasource_data_report(full_json=report)
        except ValueError as e:
            self.kismet.send_datasource_error_report(message = "Could handle JSON output")
            return False
        except Exception as e:
            self.kismet.send_datasource_error_report(message = "Could not handle output")
            return False

        return True

    def __async_radio_thread(self):
        # This function blocks forever until cancelled
        try:
            self.rtlsdr.read_samples(self.rtl_data_cb, 12, self.usb_buf_sz)
        except rtlsdr.RadioOperationalError as e:
            if not self.kismet.inSpindown():
                self.kismet.send_datasource_error_report(message = f"Error reading from RTLSDR: {e}")
        except Exception as e:
            self.kismet.send_datasource_error_report(message = f"Error reading from RTLSDR: {e}")

        # Always make sure we die
        self.kill_amr()
        self.kismet.spindown()

    def open_radio(self, rnum):
        try:
            self.rtlsdr.open_radio(rnum, self.frequency, self.rate, gain=self.opts['gain'], autogain=True, ppm=self.opts['ppm'], biastee=self.opts['biastee'])
        except Exception as e:
            return [False, f"Error opening RTLSDR device: {e.args[0]}"]

        self.rtl_thread = threading.Thread(target=self.__async_radio_thread)
        self.rtl_thread.daemon = True
        self.rtl_thread.start()

        return [True, ""]

    async def __rtl_amr_task(self):
        """
        asyncio task that consumes the output from the rtl radio
        """
        print_stderr = False

        if self.opts['debug'] is not None and self.opts['debug']:
            print_stderr = True

        try:
            while not self.kismet.kill_ioloop:
                msg = await self.message_queue.get()

                if not msg:
                    break

                if print_stderr:
                    print(msg, file=sys.stderr)

                l = json.dumps(msg)

                if not self.handle_json(l):
                    raise RuntimeError('could not send rtlamr data')
        except Exception as e:
            traceback.print_exc(file=sys.stderr)
            print("An error occurred reading from the rtlsdr; is your USB device plugged in?  Make sure that no other programs are using this rtlsdr radio.", file=sys.stderr);

            self.kismet.send_datasource_error_report(message = "Error handling AMR: {}".format(e))

        finally:
            self.kill_amr()
            self.kismet.spindown()
            return

    def rtl_data_cb(self, buf, buflen, ctx):
        if buflen == 0:
            raise RuntimeError("received empty data from rtlsdr")

        nb = np.ctypeslib.as_array(buf, shape=(buflen,)).astype(np.uint8)
        self.process(nb)
        return 

    def bch_checksum(self, buf, init = 0):
        crc = init

        for b in buf:
            crc &= 0xFFFF
            crc = crc << 8 ^ self.bch_table[crc >> 8 ^ b]

        return crc & 0xFFFF

    def cumsum(self, data, w):
        ret = np.cumsum(data)
        ret[w:] = ret[w:] - ret[:-w]
        return ret[w - 1:] / w

    def moving_average(self, data, w):
        return self.cumsum(data, w)

    def _resample_quantize(self, buf):
        # Trim trailing byte if we don't have even I/Q pairs
        if len(buf) % 2 != 0:
            buf = buf[:-1]

        # Compute the magnitude and remove the DC offset using the lookup table;
        # buf is now a real magnitude
        buf = np.add(self.square_lut[buf[::2]], self.square_lut[buf[1::2]])

        # Filter with a sub-width of the message - because we decimate AFTER
        # quantization, we window on the original symbol length
        r = self.moving_average(buf, int(self.symbol_len / 8))

        # Sliding average across the half the message width
        rm = self.moving_average(r, int(self.message_len_s * 0.5) )
        r = (r[:len(rm)] - rm)[:np.newaxis]

        # Quantize
        bits = np.where(r > 0, 1, 0)

        # Fake decimation of the bits themselves after the quanitization
        bits = bits[::self.decimation]

        return bits

    def _power_estimate(self, buf, start_bit, sz_bits):
        # Take a rough power estimate, we get it in dBFS; db relative to full scale.
        # This will get treated as dbm elsewhere in kismet which is fundamentally
        # wrong, but no more wrong than some other power measurements from other 
        # cards.  we do our best.
        bit_offt = start_bit * self.decimation
        bit_len = sz_bits * self.decimation
        powr = np.average(np.add(self.square_lut[buf[start_bit:start_bit + bit_len:2]], self.square_lut[buf[start_bit + 1:start_bit + bit_len:2]]))
        return int(10 * math.log10(powr))

    def _single_manchester(self, a, b, c, d):
        bit_p = a > b
        bit = c > d

        if bit and bit_p and c > b:
            return 1
        if bit and not bit_p and d < b:
            return 1
        if not bit and bit_p and d > b:
            return 0
        if not bit and not bit_p and c < b:
            return 0

        return None

    def corr_preamble(self, buf):
        corr = np.correlate(buf, self.search_preamble)
        return np.argmax(corr)

    def get_bits_as_int(self, buf):
        pad = len(buf) % 8
        if pad != 0:
            padbuf = np.array([0] * (8 - pad))
            buf = np.append(padbuf, buf)

        return int.from_bytes(np.packbits(buf), byteorder='big', signed=False)

    def reduce_bits(self, bits, sz):
        return bits[int(self.reduced_w / 2):(sz * self.reduced_w):self.reduced_w]

    def process(self, buf):
        # 'decimate' and quantize to a stream of bits; the bitstream returned
        # is still expanded by the sample multiplier / the bit width
        rs = self._resample_quantize(buf)

        i = 0
        while True:
            if i + len(self.scm_preamble) >= len(rs):
                break

            # Correlate for preamble; this is one of the most expensive parts of the
            # whole process
            p = self.corr_preamble(rs[i:i + (self.reduced_preamble_l * 4)])

            if p == None:
                break

            p = i + p

            if p + (24 * 8 * self.reduced_w) < len(rs):
                # Convert to one bit per symbol
                bits = self.reduce_bits(rs[p:], (14 * 8 * 2))

                i = p + self.reduced_preamble_l * 4

                # Compare the full preamble since we only match on the first 16 bits
                if not np.array_equal(bits[:self.scm_preamble_len], self.scm_preamble):
                    continue

                if len(bits) % 2 != 0:
                    bits = bits[:-1]

                errors = 0
                msgbuf = np.array([0] * int(len(bits) / 2))
                a = 0
                b = 1

                if len(bits) < (24 * 8):
                    break

                # This is the simplest way to decode manchester with no error awareness
                # but there is no huge benefit to using it, see below
                # msgbuf = np.where(bits[::2] > bits[1::2], 1, 0)

                # This isn't the simplest way to convert to manchester, but testing
                # shows almost no impact on CPU load between this and the simplest
                # non-error-checking encoding method since it's called only on the
                # resolved end-stage signal
                for ix in range(0, len(bits), 2):
                    bit = self._single_manchester(bits[ix], bits[ix+1], a, b)

                    a = bits[ix]
                    b = bits[ix+1]

                    if bit == None:
                        errors = errors + 1

                        if errors > 5:
                            break
                        else:
                            if a > b:
                                bit = 1
                            else:
                                bit = 0

                            a = 0
                            b = 1

                    msgbuf[int(ix / 2)] = bit
                    i = p + ix
               
                # SCM frame format
                # [  0 : 21 ] 21 Sync / RF Preamble 1F2A60
                # [ 21 : 23 ] 2  ID MSB
                # [ 23      ] 1  Reserved
                # [ 24 : 26 ] 2  Physical tamper
                # [ 26 : 30 ] 4  Endpoint type
                # [ 30 : 32 ] 2  Endpoint tamper
                # [ 32 : 56 ] 24 Consumption value
                # [ 56 : 80 ] 24 ID LSB
                # [ 80 : 96 ] 16 Checksum

                msgbuf = msgbuf[1:]
                bytestr = np.packbits(msgbuf);

                report_msg = {
                        "amr_scm": bytearray(bytestr).hex(),
                        "valid": False,
                        "type": "SCM",
                        }

                checksum = self.get_bits_as_int(msgbuf[80:96])

                # Anything with a checksum of 0 is summarily useless, don't even report it, it's
                # a malformed fragment
                if checksum == 0:
                    continue

                calc_checksum = self.bch_checksum(bytestr[2:10])

                # Defer power calc until we know we have something sane-ish
                pwr = self._power_estimate(buf, p, 14*8*2)

                report_msg["signal"] = pwr

                # Bounce invalid messages but report the signal anyhow
                if checksum != calc_checksum:
                    if self.opts['debug']:
                        print(report_msg)
                    self.kismet.add_task(self.message_queue.put, [report_msg])
                    continue

                # Flag as valid and start populating
                report_msg["valid"] = True

                meterid = (self.get_bits_as_int(msgbuf[21:23]) << 24)
                meterid |= self.get_bits_as_int(msgbuf[56:80])

                report_msg["meterid"] = meterid
                report_msg["metertype"] = self.get_bits_as_int(msgbuf[26:30])
                report_msg["consumption"] = self.get_bits_as_int(msgbuf[32:56])
                report_msg["phytamper"] = self.get_bits_as_int(msgbuf[24:26])
                report_msg["endptamper"] = self.get_bits_as_int(msgbuf[30:32])
                
                if self.opts['debug']:
                    print(report_msg)

                self.kismet.add_task(self.message_queue.put, [report_msg])

            else:
                break




"""
freaklabs zigbee sniffer source

(c) 2018 Mike Kershaw / Dragorn
Licensed under GPL2 or above

accepts standard source hop options
accepts additional options:

device=/path/to/serial
baud=baudrate
band=800|900|2400

Based in part on the Sensniff code from:
https://github.com/freaklabs/sensniff-freaklabs.git

Under the following license:

Copyright (c) 2012, George Oikonomou (oikonomou@users.sf.net)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
  * Neither the name of the owner nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

"""

import argparse
from datetime import datetime
import json
import os

try:
    import serial
except ImportError:
    raise ImportError("KismetCaptureFreaklabsZigbee requires python-serial, please install it!")

import struct
import subprocess
import sys
import threading
import time
import uuid

from . import kismetexternal

LINKTYPE_IEEE802_15_4_NOFCS = 230
LINKTYPE_IEEE802_15_4 = 195
NETWORK = LINKTYPE_IEEE802_15_4_NOFCS

CMD_FRAME = 0x00
CMD_CHANNEL = 0x01
CMD_GET_CHANNEL = 0x81
CMD_SET_CHANNEL = 0x82
SNIFFER_PROTO_VERSION = 1

class FreaklabException(Exception):
    pass

class SerialInputHandler(object):
    def __init__(self, port, baudrate):
        self.__sensniff_magic_legacy = struct.pack('BBBB', 0x53, 0x6E, 0x69, 0x66)
        self.__sensniff_magic = struct.pack('BBBB', 0xC1, 0x1F, 0xFE, 0x72)
        self._current_channel = -1

        try:
            self.port = serial.Serial(port = port,
                                      baudrate = baudrate,
                                      bytesize = serial.EIGHTBITS,
                                      parity = serial.PARITY_NONE,
                                      stopbits = serial.STOPBITS_ONE,
                                      xonxoff = False,
                                      rtscts = False,
                                      timeout = 0.1)
            self.port.flushInput()
            self.port.flushOutput()
        except (serial.SerialException, ValueError, IOError, OSError) as e:
            raise FreaklabException("Could not open freaklabs device: {}".format(e))

    def read_frame(self):
        try:
            # Read the magic + 1 more byte
            b = self.port.read(5)
            size = len(b)
        except (IOError, OSError) as e:
            raise FreaklabException("Error reading port: {}".format(e))

        if size == 0:
            return b
        if size < 5:
            self.port.flushInput()
            return ''

        if b[0:4] not in (self.__sensniff_magic, self.__sensniff_magic_legacy):
            # Peripheral UART output - print it
            per_out = self.port.readline().rstrip()
            return ''

        # If we reach here:
        # Next byte == 1: Proto version 1, header follows
        # Next Byte != 1 && < 128. Old proto version. Frame follows, length == the byte
        b = bytearray(b)
        if b[4] != SNIFFER_PROTO_VERSION:
            # Legacy contiki sniffer support. Will slowly fade away
            size = b[4]
            try:
                b = self.port.read(size)
            except (IOError, OSError) as e:
                raise FreaklabException("Error reading port: {}".format(e))
                return

            if len(b) != size:
                # We got the magic right but subsequent bytes did not match
                # what we expected to receive
                self.port.flushInput()
                return ''

            return b

        # If we reach here, we have a packet of proto ver SNIFFER_PROTO_VERSION
        # Read CMD and LEN
        try:
            b = self.port.read(2)

        except (IOError, OSError) as e:
            raise FreaklabException("Error reading port: {}".format(e))
            return

        if size < 2:
            self.port.flushInput()
            return ''

        b = bytearray(b)
        cmd = b[0]
        length = b[1]

        # Read the frame or command response
        b = self.port.read(length)
        if len(b) != length:
            # We got the magic right but subsequent bytes did not match
            # what we expected to receive
            self.port.flushInput()
            return ''

        # If we reach here, b holds a frame or a command response of length len
        if cmd == CMD_FRAME:
            return b

        # If we reach here, we have a command response
        b = bytearray(b)
        if cmd == CMD_CHANNEL:
             self._current_channel = b[0]
        return ''

    def __write_command(self, cmd):
        self.port.write(self.__sensniff_magic)
        self.port.write(bytearray([SNIFFER_PROTO_VERSION]))
        self.port.write(cmd)
        self.port.flush()

    def set_channel(self, channel):
        self.__write_command(bytearray([CMD_SET_CHANNEL, 1, channel]))
        # this hardware takes 150us for PLL lock and we need enough time to read the success message
        time.sleep (0.003)
        if (channel != self._current_channel):
            raise FreaklabException

    def get_channel(self):
        self.__write_command(bytearray([CMD_GET_CHANNEL]))

class KismetFreaklabsZigbee(object):
    def __init__(self):
        # Frequency map
        self.frequencies = {
            0: 868,
            1: 906,
            2: 908,
            3: 910,
            4: 912,
            5: 914,
            6: 916,
            7: 918,
            8: 920,
            9: 922,
            10: 924,
            11: 2405,
            12: 2410,
            13: 2415,
            14: 2420,
            15: 2425,
            16: 2430,
            17: 2435,
            18: 2440,
            19: 2445,
            20: 2450,
            21: 2455,
            22: 2460,
            23: 2465,
            24: 2470,
            25: 2475,
            26: 2480
        }

        self.band_map = {}
        self.band_map["800"] = ["0"]
        self.band_map["900"] = []
        self.band_map["2400"] = []
        # Freaklabs 900mhz hw can operate down to channel 0, include it in the list
        for c in range(0, 11):
            self.band_map["900"].append("{}".format(c))
        for c in range(11, 27):
            self.band_map["2400"].append("{}".format(c))

        self.defaults = {}

        self.defaults['device'] = "/dev/ttyUSB0"
        self.defaults['baudrate'] = "57600"
        self.defaults['band'] = "auto"
        self.defaults['name'] = None

        self.hop_thread = None
        self.monitor_thread = None

        self.chan_config_lock = threading.RLock()
        self.chan_config = {}
        self.chan_config['chan_pos'] = 0
        self.chan_config['hopping'] = True
        self.chan_config['channel'] = "0"
        self.chan_config['hop_channels'] = []
        self.chan_config['hop_rate'] = 1
        self.chan_config['chan_skip'] = 0
        self.chan_config['chan_offset'] = 0

        self.serialhandler = None

        parser = argparse.ArgumentParser(description='Kismet datasource to capture from Freaklabs Zigbee hardware',
                epilog='Requires Freaklabs hardware (or compatible SenSniff-based device)')
        
        parser.add_argument('--in-fd', action="store", type=int, dest="infd")
        parser.add_argument('--out-fd', action="store", type=int, dest="outfd")
        parser.add_argument('--connect', action="store", dest="connect")
        parser.add_argument("--source", action="store", dest="source")
        
        self.config = parser.parse_args()

        if not self.config.connect == None and self.config.source == None:
            print("You must specify a source with --source when connecting to a remote Kismet server")
            sys.exit(0)

        self.proberet = None

        if not self.config.source == None:
            (source, options) = kismetexternal.Datasource.parse_definition(self.config.source)

            if source == None:
                print("Could not parse the --source option; this should be a standard Kismet source definition.")
                sys.exit(0)

            self.proberet = self.datasource_probesource(source, options)

            if self.proberet == None:
                print("Could not configure local source {}, check your source options and config.")
                sys.exit(0)

            if not "success" in self.proberet:
                print("Could not configure local source {}, check your source options and config.")
                if "message" in self.proberet:
                    print(self.proberet["message"])
                sys.exit(0)

            if not self.proberet["success"]:
                print("Could not configure local source {}, check your source options and config.")
                if "message" in self.proberet:
                    print(self.proberet["message"])
                sys.exit(0)

            print("Connecting to remote server {}".format(self.config.connect))

        self.kismet = kismetexternal.Datasource(self.config.infd, self.config.outfd, remote = self.config.connect)

        self.kismet.set_configsource_cb(self.datasource_configure)
        self.kismet.set_listinterfaces_cb(self.datasource_listinterfaces)
        self.kismet.set_opensource_cb(self.datasource_opensource)
        self.kismet.set_probesource_cb(self.datasource_probesource)

        # If we're connecting remote, kick a newsource
        if self.proberet:
            print("Registering remote source {} {}".format('freaklabszigbee', self.config.source))
            self.kismet.send_datasource_newsource(self.config.source, 'freaklabszigbee', self.proberet['uuid'])

        self.kismet.start()

    def is_running(self):
        return self.kismet.is_running()

    def __start_hopping(self):
        def hop_func():
            while self.chan_config['hopping']:
                wait_usec = 1.0 / self.chan_config['hop_rate']

                try:
                    self.chan_config_lock.acquire()
                    c = int(self.chan_config['hop_channels'][self.chan_config['chan_pos'] % len(self.chan_config['hop_channels'])])
                    self.serialhandler.set_channel(c)
                except FreaklabException as e:
                    self.kismet.send_datasource_error_report(message = "Could not tune to {}: {}".format(self.chan_config['chan_pos'], e))
                finally:
                    self.chan_config_lock.release()

                self.chan_config['chan_pos'] = self.chan_config['chan_pos'] + 1

            self.hop_thread = None

        if self.hop_thread:
            return

        self.hop_thread = threading.Thread(target = hop_func)
        self.hop_thread.daemon = True
        self.hop_thread.start()

    def __detect_band(self, device):
        try:
            self.serialhandler.set_channel(2)
            self.kismet.send_message("Found band \'900MHz\' for freaklabs source on \'{}\'".format(device))
            return "900"
        except FreaklabException as e:
            True

        try:
            self.serialhandler.set_channel(13)
            self.kismet.send_message("Found band \'2.4GHz\' for freaklabs source on \'{}\'".format(device))
            return "2400"
        except FreaklabException as e:
            return "unknown"

    def __start_monitor(self):
        def mon_func():
            while self.kismet.is_running():
                try:
                    raw = self.serialhandler.read_frame()
                except FreaklabException as e:
                    self.kismet.send_datasource_error_report(message = "Error reading from zigbee device: {}".format(e))
                    break

                if len(raw) == 0:
                    continue

                packet = kismetexternal.datasource_pb2.SubPacket()
                dt = datetime.now()
                packet.time_sec = int(time.mktime(dt.timetuple()))
                packet.time_usec = int(dt.microsecond)

                packet.dlt = LINKTYPE_IEEE802_15_4_NOFCS

                packet.size = len(raw)
                packet.data = raw

                self.kismet.send_datasource_data_report(full_packet = packet)

            self.monitor_thread = None

        if self.monitor_thread:
            return

        self.monitor_thread = threading.Thread(target = mon_func)
        self.monitor_thread.daemon = True
        self.monitor_thread.start()

    # We can't really list interfaces other than to guess about serial ports which
    # seems like a bad idea; maybe we do that, eventually
    def datasource_listinterfaces(self, seqno):
        interfaces = []
        self.kismet.send_datasource_interfaces_report(seqno, interfaces)

    def __get_uuid(self, opts):
        if ('uuid' in opts):
            return opts['uuid']

        uhash = kismetexternal.Datasource.adler32("{}{}{}".format(opts['device'], opts['baudrate'], opts['name']))
        uhex = "0000{:02X}".format(uhash)

        return kismetexternal.Datasource.make_uuid("kismet_cap_freaklabs_zigbee", uhex)

    # Implement the probesource callback for the datasource api
    def datasource_probesource(self, source, options):
        ret = {}

        if not source == "freaklabs":
            return None

        opts = options
        for x in self.defaults:
            opts.setdefault(x, self.defaults[x])

        ret['uuid'] = self.__get_uuid(opts)

        try:
            SerialInputHandler(opts['device'], int(opts['baudrate']))
        except FreaklabException as e:
            ret['success'] = False
            ret['message'] = "{}".format(e)
            return ret

        ret['capture_interface'] = opts['device']
        if not opts['band'] == 'auto':
            ret['hardware'] = "freaklabs-{}".format(opts['band'])

        ret['success'] = True
        return ret

    def datasource_opensource(self, source, options):
        ret = {}

        if not source == "freaklabs":
            return None

        opts = options
        for x in self.defaults:
            opts.setdefault(x, self.defaults[x])

        ret['uuid'] = self.__get_uuid(opts)

        try:
            self.serialhandler = SerialInputHandler(opts['device'], int(opts['baudrate']))
            self.serialhandler.get_channel()
        except FreaklabException as e:
            ret['success'] = False
            ret['message'] = "{}".format(e)
            return ret

        # Launch the monitor thread
        self.__start_monitor()

        while True:
            try:
                self.serialhandler.set_channel(2)
                break
            except:
                time.sleep(0.1)

            try:
                self.serialhandler.set_channel(13)
                break
            except:
                time.sleep(0.4)


        if opts['band'] == "auto":
            opts['band'] = self.__detect_band(opts['device'])

        if opts['band'] == "unknown":
            ret['success'] = False
            ret['message'] = "Failed to auto-detect band"
            return ret

        if not opts['band'] in self.band_map:
            ret['success'] = False
            ret['message'] = "Unknown band {}".format(opts['band'])
            return ret

        band = self.band_map[opts['band']]

        ret['phy'] = LINKTYPE_IEEE802_15_4_NOFCS

        ret['channel'] = band[0]
        ret['channels'] = band

        ret['capture_interface'] = opts['device']
        ret['hardware'] = "freaklabs-{}".format(opts['band'])

        ret['success'] = True

        return ret

    def datasource_configure(self, seqno, config):
        ret = {}

        if config.HasField('channel'):
            self.chan_config_lock.acquire()
            self.chan_config['hopping'] = False
            self.chan_config['channel'] = config.channel.channel
            ret['channel'] = config.channel.channel
            self.chan_config_lock.release()
        elif config.HasField('hopping'):
            self.chan_config_lock.acquire()
            if config.hopping.HasField('rate'):
                self.chan_config['hop_rate'] = config.hopping.rate

            if len(config.hopping.channels):
                self.chan_config['hop_channels'] = []
                for c in config.hopping.channels:
                    self.chan_config['hop_channels'].append(c)

                self.chan_config['hopping'] = True

            self.chan_config_lock.release()

            # Echo its config back at it
            ret['full_hopping'] = config.hopping

        ret['success'] = True

        if self.chan_config['hopping'] and not self.hop_thread:
            self.__start_hopping()

        return ret


